library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library altera_mf;
use altera_mf.altera_mf_components.all;

entity firevid is
generic
(
	WIDTH  : natural := 640;
	HEIGHT : natural := 480
);
port
(
	tm4_glbclk0	: in std_logic;
	resetn		: in std_logic;

	-- Connect to one of the TM4's Firewire chipsets
	tm4_fir_ma	: out std_logic_vector(0 to 6);
	tm4_fir_md	: inout std_logic_vector(0 to 15);
	tm4_fir_mwrn	: out std_logic;
	tm4_fir_mcsn	: out std_logic;
	tm4_fir_mcan	: in std_logic;
	tm4_fir_tean	: in std_logic;
	tm4_fir_coldfire : out std_logic;
	tm4_fir_lendian	: out std_logic;
	tm4_fir_m8bit	: out std_logic;
	tm4_fir_mcmode	: out std_logic;
	tm4_fir_bclk	: out std_logic;
	tm4_fir_resetn	: out std_logic;
	tm4_fir_dmclk	: in std_logic;
	tm4_fir_dmd	: inout std_logic_vector(0 to 15);
	tm4_fir_dmrw	: in std_logic;
	tm4_fir_cystart	: in std_logic;
	tm4_fir_pktflag	: in std_logic;
	tm4_fir_dmpre	: in std_logic;
	tm4_fir_dmdone	: in std_logic;
	tm4_fir_dmerror	: in std_logic;
	tm4_fir_dmready	: out std_logic;

	-- Connect to instance of tm4_portmux (with proper entries added
	-- in the .ports file for FPGA1
	signals_in	: in std_logic_vector(31 downto 0);
	signals_out	: out std_logic_vector(31 downto 0);
	sigin_wanted	: out std_logic;
	sigin_ready	: in std_logic;

	-- Camera A signals
	columnA		: buffer std_logic_vector(9 downto 0);
	rowA		: buffer std_logic_vector(9 downto 0);
	redA		: out std_logic_vector(7 downto 0);
	greenA		: out std_logic_vector(7 downto 0);
	blueA		: out std_logic_vector(7 downto 0);
	datavalidA	: buffer std_logic;
	newlineA	: out std_logic;
	newframeA	: buffer std_logic;

	-- Camera B signals
	columnB		: buffer std_logic_vector(9 downto 0);
	rowB		: buffer std_logic_vector(9 downto 0);
	redB		: out std_logic_vector(7 downto 0);
	greenB		: out std_logic_vector(7 downto 0);
	blueB		: out std_logic_vector(7 downto 0);
	datavalidB	: buffer std_logic;
	newlineB	: out std_logic;
	newframeB	: buffer std_logic
);
end firevid;

architecture rtl of firevid is
	-- Wrapper between portmux and the TSB12LV32 microcontroller
	-- interface.
	component llc_iface
	port
	(
		tm4_glbclk0	: in std_logic;
		reset_n		: in std_logic;
		signals_in	: in std_logic_vector(31 downto 0);
		signals_out	: out std_logic_vector(31 downto 0);
		sigin_wanted	: out std_logic;
		sigin_ready	: in std_logic;
		tm4_fir_ma	: out std_logic_vector(0 to 6);
		tm4_fir_md	: inout std_logic_vector(0 to 15);
		tm4_fir_mwrn	: out std_logic;
		tm4_fir_mcsn	: out std_logic;
		tm4_fir_mcan	: in std_logic;
		tm4_fir_tean	: in std_logic;
		tm4_fir_coldfire : out std_logic;
		tm4_fir_lendian	: out std_logic;
		tm4_fir_m8bit	: out std_logic;
		tm4_fir_mcmode	: out std_logic;
		tm4_fir_bclk	: out std_logic;
		tm4_fir_resetn	: out std_logic
	);
	end component;
	
	component dm_fifo
	port
	(
		clock			: in	std_logic;
		reset_n			: in	std_logic;
 
		tm4_fir_dmclk		: in 	std_logic;
		tm4_fir_dmd		: inout std_logic_vector(0 to 15);
		tm4_fir_dmrw		: in 	std_logic;
		tm4_fir_cystart		: in 	std_logic;
		tm4_fir_pktflag		: in 	std_logic;
		tm4_fir_dmpre		: in 	std_logic;
		tm4_fir_dmdone		: in 	std_logic;
		tm4_fir_dmerror		: in 	std_logic;
		tm4_fir_dmready		: out 	std_logic;

		quadlet_data	: out std_logic_vector(31 downto 0);
		read_ackA	: in std_logic;
		quadlet_validA	: out std_logic;
		new_frameA	: out std_logic;
		read_ackB	: in std_logic;
		quadlet_validB	: out std_logic;
		new_frameB	: out std_logic
	);
	end component;
	
	component firevid_decoder_yuv411
	generic
	(
		WIDTH : natural;
		HEIGHT : natural
	);
	port
	(
		clock		: in	std_logic;
		reset_n		: in	std_logic;
	
		read_ack	: out	std_logic;
		quadlet_valid	: in	std_logic;
		quadlet_data	: in	std_logic_vector(31 downto 0);
		new_frame	: in	std_logic;
	
		row		: out	std_logic_vector(9 downto 0);
		column		: out	std_logic_vector(9 downto 0);
		red		: out	std_logic_vector(7 downto 0);
		green		: out	std_logic_vector(7 downto 0);
		blue		: out	std_logic_vector(7 downto 0);
		data_valid	: out	std_logic
	);
	end component;

	-- signals to/from camera fifos
	signal fifoA_data : std_logic_vector(24 downto 0);
	signal fifoA_q : std_logic_vector(24 downto 0);
	signal fifoA_wrreq : std_logic;
	signal fifoA_rdreq : std_logic;
	signal fifoA_full : std_logic;
	signal fifoA_empty : std_logic;
	signal fifoB_data : std_logic_vector(24 downto 0);
	signal fifoB_q : std_logic_vector(24 downto 0);
	signal fifoB_wrreq : std_logic;
	signal fifoB_rdreq : std_logic;
	signal fifoB_full : std_logic;
	signal fifoB_empty : std_logic;

	-- outputs of dm_fifo / inputs of video decoders
	signal read_ackA : std_logic;
	signal read_ackB : std_logic;
	signal quadlet_validA : std_logic;
	signal quadlet_validB : std_logic;
	signal quadlet_data : std_logic_vector(31 downto 0);
	signal new_frameA : std_logic;
	signal new_frameB : std_logic;

	-- outputs of video decoders
	signal row0 : std_logic_vector(9 downto 0);
	signal column0 : std_logic_vector(9 downto 0);
	signal red0 : std_logic_vector(7 downto 0);
	signal green0 : std_logic_vector(7 downto 0);
	signal blue0 : std_logic_vector(7 downto 0);
	signal data_valid0 : std_logic;
	signal new_frame0 : std_logic;
	signal row1 : std_logic_vector(9 downto 0);
	signal column1 : std_logic_vector(9 downto 0);
	signal red1 : std_logic_vector(7 downto 0);
	signal green1 : std_logic_vector(7 downto 0);
	signal blue1 : std_logic_vector(7 downto 0);
	signal data_valid1 : std_logic;
	signal new_frame1 : std_logic;

	-- signals for final generation/registering of signals
	signal is_newframeA : boolean;
	signal is_newframeB : boolean;
	signal rowA_in : std_logic_vector(9 downto 0);
	signal rowB_in : std_logic_vector(9 downto 0);
	signal columnA_in : std_logic_vector(9 downto 0);
	signal columnB_in : std_logic_vector(9 downto 0);
	signal datavalidA_in : std_logic;
	signal datavalidB_in : std_logic;
begin
	fifoA : scfifo
	generic map
	(
		LPM_WIDTH	=> 25,
		LPM_NUMWORDS	=> WIDTH*3,
		LPM_WIDTHU	=> 11,
		ALMOST_FULL_VALUE	=> WIDTH*2+1
	)
	port map
	(
		data	=> fifoA_data,
		q	=> fifoA_q,
		clock	=> tm4_glbclk0,
		wrreq	=> fifoA_wrreq,
		rdreq	=> fifoA_rdreq,
		almost_full	=> fifoA_full,
		empty	=> fifoA_empty
	);

	fifoB : scfifo
	generic map
	(
		LPM_WIDTH	=> 25,
		LPM_NUMWORDS	=> WIDTH*3,
		LPM_WIDTHU	=> 11,
		ALMOST_FULL_VALUE	=> WIDTH*2+1
	)
	port map
	(
		data	=> fifoB_data,
		q	=> fifoB_q,
		clock	=> tm4_glbclk0,
		wrreq	=> fifoB_wrreq,
		rdreq	=> fifoB_rdreq,
		almost_full	=> fifoB_full,
		empty	=> fifoB_empty
	);

	dm_interface : dm_fifo
	port map
	(
		tm4_glbclk0, resetn, tm4_fir_dmclk, tm4_fir_dmd,
		tm4_fir_dmrw, tm4_fir_cystart, tm4_fir_pktflag,
		tm4_fir_dmpre, tm4_fir_dmdone, tm4_fir_dmerror,
		tm4_fir_dmready, quadlet_data, read_ackA, quadlet_validA,
		new_frameA, read_ackB, quadlet_validB, new_frameB
	);

	decoder0 : firevid_decoder_yuv411
	generic map
	(
		WIDTH => WIDTH,
		HEIGHT => HEIGHT
	)
	port map
	(
		tm4_glbclk0, resetn, read_ackA, 
		quadlet_validA,	quadlet_data, new_frameA, 
		row0, column0, red0, green0,
		blue0, data_valid0
	);

	decoder1 : firevid_decoder_yuv411
	generic map
	(
		WIDTH => WIDTH,
		HEIGHT => HEIGHT
	)
	port map
	(
		tm4_glbclk0, resetn, read_ackB, 
		quadlet_validB, quadlet_data, new_frameB, 
		row1, column1, red1, green1,
		blue1, data_valid1
	);

	mc_interface : llc_iface
	port map
	(
		tm4_glbclk0, resetn, signals_in, signals_out,
		sigin_wanted, sigin_ready, tm4_fir_ma, tm4_fir_md,
		tm4_fir_mwrn, tm4_fir_mcsn, tm4_fir_mcan, tm4_fir_tean,
		tm4_fir_coldfire, tm4_fir_lendian, tm4_fir_m8bit,
		tm4_fir_mcmode, tm4_fir_bclk, tm4_fir_resetn
	);		

	-- Inputs to fifos
	new_frame0 <= '1' when row0 = 0 and column0 = 0 else '0';
	fifoA_data <= new_frame0 & blue0 & green0 & red0;
	fifoA_wrreq <= data_valid0;
	fifoA_rdreq <= ((not fifoA_empty) and (not fifoB_empty)) or fifoA_full;
	
	new_frame1 <= '1' when row1 = 0 and column1 = 0 else '0';
	fifoB_data <= new_frame1 & blue1 & green1 & red1;
	fifoB_wrreq <= data_valid1;
	fifoB_rdreq <= ((not fifoA_empty) and (not fifoB_empty)) or fifoB_full;
	
	process (tm4_glbclk0)
	begin
		if rising_edge(tm4_glbclk0) then
			-- If a request is made this cycle, data will be output and
			-- valid during the next cycle
			datavalidA_in <= fifoA_rdreq;
			datavalidB_in <= fifoB_rdreq;
			
			datavalidA <= datavalidA_in;
			newframeA <= fifoA_q(24);
			redA <= fifoA_q(7 downto 0);
			greenA <= fifoA_q(15 downto 8);
			blueA <= fifoA_q(23 downto 16);
			if is_newframeA then
				rowA <= (others => '0');
				columnA <= (others => '0');
			else
				rowA <= rowA_in;
				columnA <= columnA_in;
			end if;
			
			datavalidB <= datavalidB_in;
			newframeB <= fifoB_q(24);
			redB <= fifoB_q(7 downto 0);
			greenB <= fifoB_q(15 downto 8);
			blueB <= fifoB_q(23 downto 16);
			if is_newframeB then
				rowB <= (others => '0');
				columnB <= (others => '0');
			else
				rowB <= rowB_in;
				columnB <= columnB_in;
			end if;
		end if;
	end process;

	-- is_newframeA/B used during the cycle just before final output
	is_newframeA <= true when fifoA_q(24) = '1' and newframeA = '0' else false;
	is_newframeB <= true when fifoB_q(24) = '1' and newframeB = '0' else false;

	-- newlineA/B calculated based on final output signals
	newlineA <= '1' when datavalidA = '1' and columnA = 0 and rowA < HEIGHT else '0';
	newlineB <= '1' when datavalidB = '1' and columnB = 0 and rowB < HEIGHT else '0';

	process (tm4_glbclk0, resetn)
	begin
		if resetn = '0' then
			rowA_in <= (others => '0');
			columnA_in <= (others => '0');
		elsif rising_edge(tm4_glbclk0) then
			if datavalidA_in = '1' then
				if is_newframeA then
					-- If it's a new frame now, next cycle will have
					-- coordinates x=1 y=0
					columnA_in <= conv_std_logic_vector(1, 10);
					rowA_in <= conv_std_logic_vector(0, 10);
				elsif columnA_in = WIDTH-1 then
					columnA_in <= (others => '0');
					rowA_in <= rowA_in + 1;
				else
					columnA_in <= columnA_in + 1;
				end if;
			end if;
		end if;
		
		if resetn = '0' then
			rowB_in <= (others => '0');
			columnB_in <= (others => '0');
		elsif rising_edge(tm4_glbclk0) then
			if datavalidB_in = '1' then 
				if is_newframeB then
					columnB_in <= conv_std_logic_vector(1, 10);
					rowB_in <= conv_std_logic_vector(0, 10);
				elsif columnB_in = WIDTH-1 then
					columnB_in <= (others => '0');
					rowB_in <= rowB_in + 1;
				else
					columnB_in <= columnB_in + 1;
				end if;
			end if;
		end if;
	end process;
end rtl;
