#ifndef _LLCIFACE_
#define _LLCIFACE_

#ifdef __cplusplus
extern "C"
{
#endif

// Packet token: Precedes every RECEIVED firewire packet
// and describes its size, transmission speed,
// and the acknowledgement code for the request
// that generated this packet.
typedef struct
{
	int	quadlet_count;
	int	ack_code;
	int	speed;
} llc_packet_token;

// Defines a firewire packet that can be used for
// sending/receiving both requests and responses
// to firewire nodes.
typedef struct
{
	// Defined by 1394 standard
	int	spd;
	int	tlabel;
	int	rt;
	int	tcode;
	int	ext_tcode;
	int	rcode;
	int	source_id;
	int	dest_id;
	long long int	dest_offset;
	int	data_length;
	int*	data;

	// Contains the packet token for this packet
	llc_packet_token	token;
} llc_packet;

// Defines a firewire self-id packet.
typedef struct
{
	// Defined by 1394 standard
	int	phy_id;
	int	link;
	int	gap_cnt;
	int	sp;
	int	contender;
	int	pwr;
	int	p0;
	int	p1;
	int	p2;
	int	init;
	int	more;

	// In an array of these structures, this
	// defines how many elements there are in
	// the entire array (same value duplicated
	// for every member).
	int	num_pkts;
} llc_selfid_packet;


// Functions return -1 if error, 0 otherwise.
// Functions that return pointers return NULL on error.

// Lowest-level interface for reading/writing
// to TSB12LV32 registers.
int llc_init();
int llc_read(int addr);
int llc_write(int addr, int data);
int llc_write_masked(int addr, int data, int mask);
int llc_read32(int addr, int* buf);
int llc_write32(int addr, int data);

// Higher-level functions that send/recieve packets
void llc_free_packet(llc_packet* pkt);
int llc_send_packet(llc_packet* pkt);
llc_packet* llc_recv_packet();
llc_selfid_packet* llc_recv_selfid_packets();

int llc_get_node_id();
int llc_packet_available();
int llc_get_last_ack();

#ifdef __cplusplus
}
#endif

#endif
